EXPORTS	TIFFAccessTagMethods
	TIFFCIELabToRGBInit
	TIFFCIELabToXYZ
	TIFFCheckTile
	TIFFCheckpointDirectory
	TIFFCleanup
	TIFFClientOpen
	TIFFClientdata
	TIFFClose
	TIFFComputeStrip
	TIFFComputeTile
	TIFFCreateCustomDirectory
	TIFFCreateDirectory
	TIFFCreateEXIFDirectory
	TIFFCreateGPSDirectory
	TIFFCurrentDirOffset
	TIFFCurrentDirectory
	TIFFCurrentRow
	TIFFCurrentStrip
	TIFFCurrentTile
	TIFFDataWidth
	TIFFDefaultStripSize
	TIFFDefaultTileSize
	TIFFDeferStrileArrayWriting
	TIFFError
	TIFFErrorExt
	TIFFFdOpen
	TIFFFieldDataType
	TIFFFieldName
	TIFFFieldPassCount
	TIFFFieldReadCount
	TIFFFieldIsAnonymous
	TIFFFieldTag
	TIFFFieldWithName
	TIFFFieldWithTag
	TIFFFieldWriteCount
	TIFFFieldSetGetSize
	TIFFFieldSetGetCountSize
	TIFFFileName
	TIFFFileno
	TIFFFindCODEC
	TIFFFindField
	TIFFFlush
	TIFFFlushData
	TIFFForceStrileArrayWriting
	TIFFFreeDirectory
	TIFFGetBitRevTable
	TIFFGetClientInfo
	TIFFGetCloseProc
	TIFFGetConfiguredCODECs
	TIFFGetField
	TIFFGetFieldDefaulted
	TIFFGetMapFileProc
	TIFFGetMode
	TIFFGetReadProc
	TIFFGetSeekProc
	TIFFGetSizeProc
	TIFFGetStrileByteCount
	TIFFGetStrileByteCountWithErr
	TIFFGetStrileOffset
	TIFFGetStrileOffsetWithErr
	TIFFGetTagListCount
	TIFFGetTagListEntry
	TIFFGetUnmapFileProc
	TIFFGetVersion
	TIFFGetWriteProc
	TIFFIsBigEndian
	TIFFIsBigTIFF
	TIFFIsByteSwapped
	TIFFIsCODECConfigured
	TIFFIsMSB2LSB
	TIFFIsTiled
	TIFFIsUpSampled
	TIFFLastDirectory
	TIFFMergeFieldInfo
	TIFFNumberOfDirectories
	TIFFNumberOfStrips
	TIFFNumberOfTiles
	TIFFOpen
	TIFFOpenW
	TIFFPrintDirectory
	TIFFRGBAImageBegin
	TIFFRGBAImageEnd
	TIFFRGBAImageGet
	TIFFRGBAImageOK
	TIFFRasterScanlineSize
	TIFFRasterScanlineSize64
	TIFFRawStripSize
	TIFFRawStripSize64
	TIFFReadBufferSetup
	TIFFReadCustomDirectory
	TIFFReadDirectory
	TIFFReadEXIFDirectory
	TIFFReadGPSDirectory
	TIFFReadEncodedStrip
	TIFFReadEncodedTile
	TIFFReadFromUserBuffer
	TIFFReadRGBAImage
	TIFFReadRGBAImageOriented
	TIFFReadRGBAStrip
	TIFFReadRGBAStripExt
	TIFFReadRGBATile
	TIFFReadRGBATileExt
	TIFFReadRawStrip
	TIFFReadRawTile
	TIFFReadScanline
	TIFFReadTile
	TIFFRegisterCODEC
	TIFFReverseBits
	TIFFRewriteDirectory
	TIFFScanlineSize
	TIFFScanlineSize64
	TIFFSetClientInfo
	TIFFSetClientdata
	TIFFSetCompressionScheme
	TIFFSetDirectory
	TIFFSetErrorHandler
	TIFFSetErrorHandlerExt
	TIFFSetField
	TIFFSetFileName
	TIFFSetFileno
	TIFFSetMode
	TIFFSetSubDirectory
	TIFFSetTagExtender
	TIFFSetWarningHandler
	TIFFSetWarningHandlerExt
	TIFFSetWriteOffset
	TIFFSetupStrips
	TIFFStripSize
	TIFFStripSize64
	TIFFSwabArrayOfDouble
	TIFFSwabArrayOfFloat
	TIFFSwabArrayOfLong
	TIFFSwabArrayOfLong8
	TIFFSwabArrayOfShort
	TIFFSwabArrayOfTriples
	TIFFSwabDouble
	TIFFSwabFloat
	TIFFSwabLong
	TIFFSwabLong8
	TIFFSwabShort
	TIFFTileRowSize
	TIFFTileRowSize64
	TIFFTileSize
	TIFFTileSize64
	TIFFUnRegisterCODEC
	TIFFUnlinkDirectory
	TIFFUnsetField
	TIFFVGetField
	TIFFVGetFieldDefaulted
	TIFFVSetField
	TIFFVStripSize
	TIFFVStripSize64
	TIFFVTileSize
	TIFFVTileSize64
	TIFFWarning
	TIFFWarningExt
	TIFFWriteBufferSetup
	TIFFWriteCheck
	TIFFWriteCustomDirectory
	TIFFWriteDirectory
	TIFFWriteEncodedStrip
	TIFFWriteEncodedTile
	TIFFWriteRawStrip
	TIFFWriteRawTile
	TIFFWriteScanline
	TIFFWriteTile
	TIFFXYZToRGB
	TIFFYCbCrToRGBInit
	TIFFYCbCrtoRGB
	_TIFFCheckMalloc
	_TIFFCheckRealloc
	_TIFFRewriteField
	_TIFFfree
	_TIFFmalloc
	_TIFFmemcmp
	_TIFFmemcpy
	_TIFFmemset
	_TIFFrealloc
	_TIFFMultiply32
	_TIFFMultiply64
	_TIFFGetExifFields
	_TIFFGetGpsFields

